(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(global = global || self, factory(global['async-bulk-map'] = {}));
}(this, function (exports) { 'use strict';

	// A type of promise-like that resolves synchronously and supports only one observer
	const _Pact = /*#__PURE__*/(function() {
		function _Pact() {}
		_Pact.prototype.then = function(onFulfilled, onRejected) {
			const result = new _Pact();
			const state = this.s;
			if (state) {
				const callback = state & 1 ? onFulfilled : onRejected;
				if (callback) {
					try {
						_settle(result, 1, callback(this.v));
					} catch (e) {
						_settle(result, 2, e);
					}
					return result;
				} else {
					return this;
				}
			}
			this.o = function(_this) {
				try {
					const value = _this.v;
					if (_this.s & 1) {
						_settle(result, 1, onFulfilled ? onFulfilled(value) : value);
					} else if (onRejected) {
						_settle(result, 1, onRejected(value));
					} else {
						_settle(result, 2, value);
					}
				} catch (e) {
					_settle(result, 2, e);
				}
			};
			return result;
		};
		return _Pact;
	})();

	// Settles a pact synchronously
	function _settle(pact, state, value) {
		if (!pact.s) {
			if (value instanceof _Pact) {
				if (value.s) {
					if (state & 1) {
						state = value.s;
					}
					value = value.v;
				} else {
					value.o = _settle.bind(null, pact, state);
					return;
				}
			}
			if (value && value.then) {
				value.then(_settle.bind(null, pact, state), _settle.bind(null, pact, 2));
				return;
			}
			pact.s = state;
			pact.v = value;
			const observer = pact.o;
			if (observer) {
				observer(pact);
			}
		}
	}

	function _isSettledPact(thenable) {
		return thenable instanceof _Pact && thenable.s & 1;
	}

	// Asynchronously iterate through an object that has a length property, passing the index as the first argument to the callback (even as the length property changes)
	function _forTo(array, body, check) {
		var i = -1, pact, reject;
		function _cycle(result) {
			try {
				while (++i < array.length && (!check || !check())) {
					result = body(i);
					if (result && result.then) {
						if (_isSettledPact(result)) {
							result = result.v;
						} else {
							result.then(_cycle, reject || (reject = _settle.bind(null, pact = new _Pact(), 2)));
							return;
						}
					}
				}
				if (pact) {
					_settle(pact, 1, result);
				} else {
					pact = result;
				}
			} catch (e) {
				_settle(pact || (pact = new Pact()), 2, e);
			}
		}
		_cycle();
		return pact;
	}

	const _iteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.iterator || (Symbol.iterator = Symbol("Symbol.iterator"))) : "@@iterator";

	// Asynchronously iterate through an object's values
	// Uses for...of if the runtime supports it, otherwise iterates until length on a copy
	function _forOf(target, body, check) {
		if (typeof target[_iteratorSymbol] === "function") {
			var iterator = target[_iteratorSymbol](), step, pact, reject;
			function _cycle(result) {
				try {
					while (!(step = iterator.next()).done && (!check || !check())) {
						result = body(step.value);
						if (result && result.then) {
							if (_isSettledPact(result)) {
								result = result.v;
							} else {
								result.then(_cycle, reject || (reject = _settle.bind(null, pact = new _Pact(), 2)));
								return;
							}
						}
					}
					if (pact) {
						_settle(pact, 1, result);
					} else {
						pact = result;
					}
				} catch (e) {
					_settle(pact || (pact = new Pact()), 2, e);
				}
			}
			_cycle();
			if (iterator.return) {
				var _fixup = function(value) {
					try {
						if (!step.done) {
							iterator.return();
						}
					} catch(e) {
					}
					return value;
				};
				if (pact && pact.then) {
					return pact.then(_fixup, function(e) {
						throw _fixup(e);
					});
				}
				_fixup();
			}
			return pact;
		}
		// No support for Symbol.iterator
		if (!("length" in target)) {
			throw new TypeError("Object is not iterable");
		}
		// Handle live collections properly
		var values = [];
		for (var i = 0; i < target.length; i++) {
			values.push(target[i]);
		}
		return _forTo(values, function(i) { return body(values[i]); }, check);
	}

	const _asyncIteratorSymbol = /*#__PURE__*/ typeof Symbol !== "undefined" ? (Symbol.asyncIterator || (Symbol.asyncIterator = Symbol("Symbol.asyncIterator"))) : "@@asyncIterator";

	// Asynchronously implement a generic for loop
	function _for(test, update, body) {
		var stage;
		for (;;) {
			var shouldContinue = test();
			if (_isSettledPact(shouldContinue)) {
				shouldContinue = shouldContinue.v;
			}
			if (!shouldContinue) {
				return result;
			}
			if (shouldContinue.then) {
				stage = 0;
				break;
			}
			var result = body();
			if (result && result.then) {
				if (_isSettledPact(result)) {
					result = result.s;
				} else {
					stage = 1;
					break;
				}
			}
			if (update) {
				var updateValue = update();
				if (updateValue && updateValue.then && !_isSettledPact(updateValue)) {
					stage = 2;
					break;
				}
			}
		}
		var pact = new _Pact();
		var reject = _settle.bind(null, pact, 2);
		(stage === 0 ? shouldContinue.then(_resumeAfterTest) : stage === 1 ? result.then(_resumeAfterBody) : updateValue.then(_resumeAfterUpdate)).then(void 0, reject);
		return pact;
		function _resumeAfterBody(value) {
			result = value;
			do {
				if (update) {
					updateValue = update();
					if (updateValue && updateValue.then && !_isSettledPact(updateValue)) {
						updateValue.then(_resumeAfterUpdate).then(void 0, reject);
						return;
					}
				}
				shouldContinue = test();
				if (!shouldContinue || (_isSettledPact(shouldContinue) && !shouldContinue.v)) {
					_settle(pact, 1, result);
					return;
				}
				if (shouldContinue.then) {
					shouldContinue.then(_resumeAfterTest).then(void 0, reject);
					return;
				}
				result = body();
				if (_isSettledPact(result)) {
					result = result.v;
				}
			} while (!result || !result.then);
			result.then(_resumeAfterBody).then(void 0, reject);
		}
		function _resumeAfterTest(shouldContinue) {
			if (shouldContinue) {
				result = body();
				if (result && result.then) {
					result.then(_resumeAfterBody).then(void 0, reject);
				} else {
					_resumeAfterBody(result);
				}
			} else {
				_settle(pact, 1, result);
			}
		}
		function _resumeAfterUpdate() {
			if (shouldContinue = test()) {
				if (shouldContinue.then) {
					shouldContinue.then(_resumeAfterTest).then(void 0, reject);
				} else {
					_resumeAfterTest(shouldContinue);
				}
			} else {
				_settle(pact, 1, result);
			}
		}
	}

	var asyncBulkMap = function asyncBulkMap(arr, bulkSize, actions) {
	  try {
	    var _temp7 = function _temp7(_actions$beforeAll) {
	      function _temp5() {
	        function _temp3(_actions$afterAll) {
	          return _actions$afterAll || res;
	        }

	        var _actions$afterAll2 = actions.afterAll;
	        return _actions$afterAll2 ? Promise.resolve(actions.afterAll(res)).then(_temp3) : _temp3(_actions$afterAll2);
	      }

	      arr = _actions$beforeAll || arr;
	      var res = [];
	      var i = 0;

	      var _temp4 = _for(function () {
	        return i < Math.max(arr.length / bulkSize, 1);
	      }, function () {
	        return i++;
	      }, function () {
	        function _temp2(_actions$beforeBulk) {
	          bulk = _actions$beforeBulk || bulk;
	          return Promise.resolve(Promise.all(bulk.map(function (item) {
	            return actions.map(item);
	          }))).then(function (bulkRes) {
	            function _temp(_actions$afterBulk) {
	              bulkRes = _actions$afterBulk || bulkRes;
	              res = res.concat(bulkRes);
	            }

	            var _actions$afterBulk2 = actions.afterBulk;
	            return _actions$afterBulk2 ? Promise.resolve(actions.afterBulk(bulkRes)).then(_temp) : _temp(_actions$afterBulk2);
	          });
	        }

	        var bulk = arr.slice(i * bulkSize, i * bulkSize + bulkSize);
	        var _actions$beforeBulk2 = actions.beforeBulk;
	        return _actions$beforeBulk2 ? Promise.resolve(actions.beforeBulk(bulk, i)).then(_temp2) : _temp2(_actions$beforeBulk2);
	      });

	      return _temp4 && _temp4.then ? _temp4.then(_temp5) : _temp5(_temp4);
	    };

	    var _actions$beforeAll3 = actions.beforeAll;
	    return Promise.resolve(_actions$beforeAll3 ? Promise.resolve(actions.beforeAll(arr)).then(_temp7) : _temp7(_actions$beforeAll3));
	  } catch (e) {
	    return Promise.reject(e);
	  }
	};

	function createPipe(pipeSrc) {
	  return createFunctionsPipe([pipeSrc]);
	}

	function createFunctionsPipe(pipeFns) {
	  return Object.assign(function () {
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }

	    try {
	      var res = args;

	      var _temp2 = _forOf(pipeFns, function (currFn) {
	        return Promise.resolve(currFn.apply(void 0, res || args)).then(function (_currFn) {
	          res = _currFn;
	        });
	      });

	      return Promise.resolve(_temp2 && _temp2.then ? _temp2.then(function () {
	        return res;
	      }) : res);
	    } catch (e) {
	      return Promise.reject(e);
	    }
	  }, {
	    pipe: function pipe(nextFn) {
	      pipeFns.push(nextFn);
	      return this;
	    },
	    then: function then(nextFn) {
	      return createFunctionsPipe(pipeFns.concat(nextFn));
	    }
	  });
	}

	function logBulk(bulk, bulkIndex) {
	  console.log("bulk index: " + bulkIndex + " (" + bulk.length + " entries)");
	}
	function createDelay(timeout) {
	  return function () {
	    return new Promise(function (resolve) {
	      return setTimeout(resolve, timeout);
	    });
	  };
	}

	exports.asyncBulkMap = asyncBulkMap;
	exports.createDelay = createDelay;
	exports.createPipe = createPipe;
	exports.logBulk = logBulk;

}));
//# sourceMappingURL=async-bulk-map.umd.development.js.map
